@@ -75,9 +75,11 @@ send_response_in_thread_func (GTask        *task,
   guint response;
   GVariant *options;
   gboolean writable = TRUE;
+  gboolean directory = TRUE;
   const char **uris;
   GVariant *choices;
   gboolean for_save;
+  GVariant *current_filter;
 
   g_variant_builder_init (&results, G_VARIANT_TYPE_VARDICT);
   g_variant_builder_init (&ruris, G_VARIANT_TYPE_STRING_ARRAY);
@@ -85,6 +87,7 @@ send_response_in_thread_func (GTask        *task,
   REQUEST_AUTOLOCK (request);
 
   for_save = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (request), "for-save"));
+  directory = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (request), "directory"));
   response = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (request), "response"));
   options = (GVariant *)g_object_get_data (G_OBJECT (request), "options");
 
@@ -98,6 +101,10 @@ send_response_in_thread_func (GTask        *task,
   if (choices)
     g_variant_builder_add (&results, "{sv}", "choices", choices);
 
+  current_filter = g_variant_lookup_value (options, "current_filter", G_VARIANT_TYPE ("(sa(us))"));
+  if (current_filter)
+    g_variant_builder_add (&results, "{sv}", "current_filter", current_filter);
+
   if (g_variant_lookup (options, "uris", "^a&s", &uris))
     {
       int i;
@@ -107,7 +114,7 @@ send_response_in_thread_func (GTask        *task,
           g_autofree char *ruri = NULL;
           g_autoptr(GError) error = NULL;
 
-          ruri = register_document (uris[i], xdp_app_info_get_id (request->app_info), for_save, writable, &error);
+          ruri = register_document (uris[i], xdp_app_info_get_id (request->app_info), for_save, writable, directory, &error);
           if (ruri == NULL)
             {
               g_warning ("Failed to register %s: %s", uris[i], error->message);
@@ -439,6 +446,7 @@ static XdpOptionKey open_file_options[] = {
   { "accept_label", G_VARIANT_TYPE_STRING, NULL },
   { "modal", G_VARIANT_TYPE_BOOLEAN, NULL },
   { "multiple", G_VARIANT_TYPE_BOOLEAN, NULL },
+  { "directory", G_VARIANT_TYPE_BOOLEAN, NULL },
   { "filters", (const GVariantType *)"a(sa(us))", validate_filters },
   { "current_filter", (const GVariantType *)"(sa(us))", validate_current_filter },
   { "choices", (const GVariantType *)"a(ssa(ss)s)", validate_choices }
@@ -456,6 +464,7 @@ handle_open_file (XdpFileChooser *object,
   g_autoptr(GError) error = NULL;
   g_autoptr(XdpImplRequest) impl_request = NULL;
   GVariantBuilder options;
+  g_autoptr(GVariant) dir_option = NULL;
 
   g_debug ("Handling OpenFile");
 
@@ -481,6 +490,12 @@ handle_open_file (XdpFileChooser *object,
       return TRUE;
     }
 
+  dir_option = g_variant_lookup_value (arg_options,
+                                       "directory",
+                                       G_VARIANT_TYPE_BOOLEAN);
+  if (dir_option && g_variant_get_boolean (dir_option))
+    g_object_set_data (G_OBJECT (request), "directory", GINT_TO_POINTER (TRUE));
+
   request_set_impl_request (request, impl_request);
   request_export (request, g_dbus_method_invocation_get_connection (invocation));
 
@@ -720,7 +735,7 @@ file_chooser_iface_init (XdpFileChooserIface *iface)
 static void
 file_chooser_init (FileChooser *fc)
 {
-  xdp_file_chooser_set_version (XDP_FILE_CHOOSER (fc), 2);
+  xdp_file_chooser_set_version (XDP_FILE_CHOOSER (fc), 3);
 }
 
 static void
