@@ -95,7 +95,7 @@ send_response_in_thread_func (GTask *task,
           goto out;
         }
 
-      ruri = register_document (uri, xdp_app_info_get_id (request->app_info), FALSE, FALSE, &error);
+      ruri = register_document (uri, xdp_app_info_get_id (request->app_info), FALSE, FALSE, FALSE, &error);
       if (ruri == NULL)
         g_warning ("Failed to register %s: %s", uri, error->message);
       else
