@@ -60,6 +60,11 @@ gboolean    xdg_app_info_map_pids        (XdpAppInfo  *app_info,
 					  pid_t       *pids,
 					  guint        n_pids,
 					  GError     **error);
+gboolean    xdg_app_info_pidfds_to_pids (XdpAppInfo  *app_info,
+                                         const int   *fds,
+                                         pid_t       *pids,
+                                         gint         count,
+                                         GError     **error);
 char *      xdp_app_info_get_path_for_fd (XdpAppInfo  *app_info,
                                           int          fd,
                                           int          require_st_mode,
@@ -168,3 +173,20 @@ gboolean xdp_spawnv     (GFile                *dir,
 char * xdp_canonicalize_filename (const char *path);
 gboolean  xdp_has_path_prefix (const char *str,
                                const char *prefix);
+
+#if !GLIB_CHECK_VERSION (2, 58, 0)
+static inline gboolean
+g_hash_table_steal_extended (GHashTable    *hash_table,
+                             gconstpointer  lookup_key,
+                             gpointer      *stolen_key,
+                             gpointer      *stolen_value)
+{
+  if (g_hash_table_lookup_extended (hash_table, lookup_key, stolen_key, stolen_value))
+    {
+      g_hash_table_steal (hash_table, lookup_key);
+      return TRUE;
+    }
+  else
+      return FALSE;
+}
+#endif
